"""
This module contains PartSeg function used for calculate in batch processing

.. graphviz::

   digraph foo {
      "CalculationManager" -> "BatchManager" -> "BatchWorker"[arrowhead="crow"];
   }
"""
import logging
import threading
import typing
from collections import OrderedDict, defaultdict
from enum import Enum
from os import path
from queue import Queue

import numpy as np
import pandas as pd
import tifffile

from PartSegCore.algorithm_describe_base import SegmentationProfile
from PartSegCore.analysis.algorithm_description import analysis_algorithm_dict
from PartSegCore.analysis.calculation_plan import (
    MaskMapper,
    MaskUse,
    MaskCreate,
    Save,
    Operations,
    FileCalculation,
    MaskIntersection,
    MaskSum,
    get_save_path,
    MeasurementCalculate,
    BaseCalculation,
    Calculation,
    RootType,
    CalculationTree,
)
from PartSegCore.analysis.io_utils import ProjectTuple
from PartSegCore.analysis.load_functions import LoadMaskSegmentation, LoadProject, load_dict
from PartSegCore.analysis.measurement_base import AreaType, PerComponent
from PartSegCore.analysis.measurement_calculation import MeasurementResult
from PartSegCore.analysis.save_functions import save_dict
from PartSegCore.mask_create import calculate_mask
from PartSegCore.segmentation.algorithm_base import report_empty_fun, SegmentationAlgorithm
from PartSegImage import Image, TiffImageReader
from .parallel_backend import BatchManager
from ...io_utils import WrongFileTypeException, HistoryElement


def do_calculation(file_path: str, calculation: BaseCalculation):
    """
    Main function which will be used for run calculation.
    It create :py:class:`.CalculationProcess` and call it method
    :py:meth:`.CalculationProcess.do_calculation`

    :param file_path: path to file which should be processed
    :param calculation: calculation description
    """
    calc = CalculationProcess()
    return calc.do_calculation(FileCalculation(file_path, calculation))


class CalculationProcess:
    """
    Main class to calculate PartSeg calculation plan
    """

    def __init__(self):
        self.reused_mask = set()
        self.mask_dict = dict()
        self.calculation = None
        self.measurement = []
        self.image: typing.Optional[Image] = None
        self.segmentation: typing.Optional[np.ndarray] = None
        self.full_segmentation: typing.Optional[np.ndarray] = None
        self.mask: typing.Optional[np.ndarray] = None
        self.history: typing.List[HistoryElement] = []
        self.algorithm_parameters: dict = {}
        self.cleaned_channel: typing.Optional[np.ndarray] = None
        self.results = []

    def do_calculation(self, calculation: FileCalculation):
        """
        Main function for calculation process

        :param calculation: calculation to do.
        :return:
        """
        self.calculation = calculation
        self.reused_mask = calculation.calculation_plan.get_reused_mask()
        self.mask_dict = {}
        self.measurement = []
        self.results = []
        operation = calculation.calculation_plan.execution_tree.operation
        ext = path.splitext(calculation.file_path)[1]
        metadata = {"default_spacing": calculation.voxel_size}
        if operation == RootType.Image:
            for load_class in load_dict.values():
                if load_class.partial() or load_class.number_of_files() != 1:
                    continue
                if ext in load_class.get_extensions():
                    projects = load_class.load([calculation.file_path], metadata=metadata)
                    break
            else:
                raise ValueError("File type not supported")
        elif operation == RootType.Project:
            projects = LoadProject.load([calculation.file_path], metadata=metadata)
        else:  # operation == RootType.Mask_project
            try:
                projects = LoadProject.load([calculation.file_path], metadata=metadata)
            except (KeyError, WrongFileTypeException):
                # TODO identify exceptions
                projects = LoadMaskSegmentation.load([calculation.file_path], metadata=metadata)

        if isinstance(projects, ProjectTuple):
            projects = [projects]
        for project in projects:
            project: ProjectTuple
            self.image = project.image
            if operation == RootType.Mask_project:
                self.mask = project.mask[0]
            if operation == RootType.Project:
                self.mask = project.mask[0]
                self.segmentation = project.segmentation
                self.full_segmentation = project.full_segmentation
                self.history = project.history
                self.algorithm_parameters = project.algorithm_parameters

            self.iterate_over(calculation.calculation_plan.execution_tree)
            for el in self.measurement:
                el.set_filename(path.relpath(project.image.file_path, calculation.base_prefix))
            self.results.append((path.relpath(project.image.file_path, calculation.base_prefix), self.measurement))
            self.measurement = []
        return self.results

    def iterate_over(self, node: typing.Union[CalculationTree, typing.List[CalculationTree]]):
        """
        Execute calculation on node children or list oof nodes

        :type node: CalculationTree
        :param node:
        :return:
        """
        if isinstance(node, CalculationTree):
            node = node.children
        for el in node:
            self.recursive_calculation(el)

    def step_load_mask(self, operation: MaskMapper, children: typing.List[CalculationTree]):
        """
        Load mask using mask mapper (mask can be defined with suffix, substitution, or file with mapping saved,
        then iterate over ``children`` nodes.

        :param MaskMapper operation: operation to perform
        :param typing.List[CalculationTree] children: list of nodes to iterate over with applied mask
        """
        mask_path = operation.get_mask_path(self.calculation.file_path)
        if mask_path == "":
            raise ValueError("Empty path to mask.")
        with tifffile.TiffFile(mask_path) as mask_file:
            mask = mask_file.asarray()
            mask = TiffImageReader.update_array_shape(mask, mask_file.series[0].axes)[..., 0]
        mask = (mask > 0).astype(np.uint8)
        try:
            mask = self.image.fit_array_to_image(mask)[0]
            # TODO fix this time bug fix
        except ValueError:
            raise ValueError("Mask do not fit to given image")
        old_mask = self.mask
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_segmentation(self, operation: SegmentationProfile, children: typing.List[CalculationTree]):
        """
        Perform segmentation and iterate over ``children`` nodes

        :param SegmentationProfile operation: Specification of segmentation operation
        :param typing.List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        segmentation_class = analysis_algorithm_dict.get(operation.algorithm, None)
        if segmentation_class is None:
            raise ValueError(f"Segmentation class {operation.algorithm} do not found")
        segmentation_algorithm = segmentation_class()
        segmentation_algorithm.set_image(self.image)
        segmentation_algorithm.set_mask(self.mask)
        segmentation_algorithm.set_parameters(**operation.values)
        result = segmentation_algorithm.calculation_run(report_empty_fun)
        backup_data = self.segmentation, self.full_segmentation, self.cleaned_channel, self.algorithm_parameters
        self.segmentation = result.segmentation
        self.full_segmentation = result.full_segmentation
        self.cleaned_channel = result.cleaned_channel
        self.algorithm_parameters = {"algorithm_name": operation.algorithm, "values": operation.values}
        self.iterate_over(children)
        self.segmentation, self.full_segmentation, self.cleaned_channel, self.algorithm_parameters = backup_data

    def step_mask_use(self, operation: MaskUse, children: typing.List[CalculationTree]):
        """
        use already defined mask and iterate over ``children`` nodes

        :param MaskUse operation:
        :param typing.List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        old_mask = self.mask
        mask = self.mask_dict[operation.name]
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_mask_operation(
        self, operation: typing.Union[MaskSum, MaskIntersection], children: typing.List[CalculationTree]
    ):
        """
        Generate new mask by sum or intersection of existing and iterate over ``children`` nodes

        :param typing.Union[MaskSum, MaskIntersection] operation: mask operation to perform
        :param typing.List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        old_mask = self.mask
        mask1 = self.mask_dict[operation.mask1]
        mask2 = self.mask_dict[operation.mask2]
        if isinstance(operation, MaskSum):
            mask = np.logical_or(mask1, mask2).astype(np.uint8)
        else:
            mask = np.logical_and(mask1, mask2).astype(np.uint8)
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_save(self, operation: Save):
        """
        Perform save operation selected in plan.

        :param Save operation: save definition
        """
        save_class = save_dict[operation.algorithm]
        project_tuple = ProjectTuple(
            file_path="",
            image=self.image,
            segmentation=self.segmentation,
            full_segmentation=self.full_segmentation,
            mask=self.mask,
            history=self.history,
            algorithm_parameters=self.algorithm_parameters,
        )
        save_path = get_save_path(operation, self.calculation)
        save_class.save(save_path, project_tuple, operation.values)

    def step_mask_create(self, operation: MaskCreate, children: typing.List[CalculationTree]):
        """
        Create mask from current segmentation state using definition

        :param MaskCreate operation: mask create description.
        :param typing.List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        mask = calculate_mask(operation.mask_property, self.segmentation, self.mask, self.image.spacing)
        if operation.name in self.reused_mask:
            self.mask_dict[operation.name] = mask
        history_element = HistoryElement.create(
            self.segmentation, self.full_segmentation, self.mask, self.algorithm_parameters, operation.mask_property,
        )
        backup = self.mask, self.history
        self.mask = mask
        self.history.append(history_element)
        self.iterate_over(children)
        self.mask, self.history = backup

    def step_measurement(self, operation: MeasurementCalculate):
        """
        Calculate measurement defined in current operation.

        :param MeasurementCalculate operation: definition of measurement to calculate
        """
        channel = operation.channel
        if channel == -1:
            segmentation_class: typing.Type[SegmentationAlgorithm] = analysis_algorithm_dict.get(
                self.algorithm_parameters["algorithm_name"], None
            )
            if segmentation_class is None:
                raise ValueError(f"Segmentation class {self.algorithm_parameters['algorithm_name']} do not found")
            channel = self.algorithm_parameters["values"][segmentation_class.get_channel_parameter_name()]

        image_channel = self.image.get_channel(channel)
        # FIXME use additional information
        measurement = operation.statistic_profile.calculate(
            image_channel, self.segmentation, self.full_segmentation, self.mask, self.image.spacing, operation.units,
        )
        self.measurement.append(measurement)

    def recursive_calculation(self, node: CalculationTree):
        """
        Identify node type and then call proper `step_*` function

        :param CalculationTree node: Node to be proceed
        """
        if isinstance(node.operation, MaskMapper):
            self.step_load_mask(node.operation, node.children)
        elif isinstance(node.operation, SegmentationProfile):
            self.step_segmentation(node.operation, node.children)
        elif isinstance(node.operation, MaskUse):
            self.step_mask_use(node.operation, node.children)
        elif isinstance(node.operation, (MaskSum, MaskIntersection)):
            self.step_mask_operation(node.operation, node.children)
        elif isinstance(node.operation, Save):
            self.step_save(node.operation)
        elif isinstance(node.operation, MaskCreate):
            self.step_mask_create(node.operation, node.children)
        elif isinstance(node.operation, Operations):
            # backward compatibility
            self.iterate_over(node)
        elif isinstance(node.operation, MeasurementCalculate):
            self.step_measurement(node.operation)
        else:
            raise ValueError("Unknown operation {} {}".format(type(node.operation), node.operation))


class ResponseData(typing.NamedTuple):
    path_to_file: str
    values: typing.List[MeasurementResult]


class CalculationManager:
    """
    This class manage batch processing in PartSeg.

    """

    def __init__(self):
        self.batch_manager = BatchManager()
        self.calculation_queue = Queue()
        self.calculation_dict = OrderedDict()
        self.calculation_sizes = []
        self.calculation_size = 0
        self.calculation_done = 0
        self.counter_dict = OrderedDict()
        self.errors_list = []
        self.sheet_name = defaultdict(set)
        self.writer = DataWriter()

    def is_valid_sheet_name(self, excel_path, sheet_name):
        return sheet_name not in self.sheet_name[excel_path]

    def add_calculation(self, calculation: Calculation):
        """
        :param calculation: alculation
        :return:
        """
        self.sheet_name[calculation.measurement_file_path].add(calculation.sheet_name)
        self.calculation_dict[calculation.uuid] = calculation, calculation.calculation_plan.get_measurements()
        self.counter_dict[calculation.uuid] = 0
        size = len(calculation.file_list)
        self.calculation_sizes.append(size)
        self.calculation_size += size
        self.batch_manager.add_work(calculation.file_list, calculation.get_base_calculation(), do_calculation)
        self.writer.add_data_part(calculation)

    @property
    def has_work(self):
        return self.batch_manager.has_work or not self.writer.writing_finished()

    def set_number_of_workers(self, val):
        logging.debug("Number off process {}".format(val))
        self.batch_manager.set_number_of_process(val)

    def get_results(self):
        responses = self.batch_manager.get_result()
        new_errors = []
        for uuid, result_list in responses:
            self.calculation_done += 1
            self.counter_dict[uuid] += 1
            calculation = self.calculation_dict[uuid][0]
            for el in result_list:
                if isinstance(el, tuple) and isinstance(el[0], Exception):
                    self.errors_list.append(el)
                    new_errors.append(el)
                else:
                    if not isinstance(el, tuple):
                        raise ValueError(f"el should be tuple. It is {type(el)}")
                    data = ResponseData._make(el)
                    errors = self.writer.add_result(data, calculation)
                    for err in errors:
                        new_errors.append(err)
                if self.counter_dict[uuid] == len(calculation.file_list):
                    errors = self.writer.calculation_finished(calculation)
                    for err in errors:
                        new_errors.append(err)
        return new_errors, self.calculation_done, zip(self.counter_dict.values(), self.calculation_sizes)


class FileType(Enum):
    excel_xlsx_file = 1
    excel_xls_file = 2
    text_file = 3


class SheetData(object):
    def __init__(self, name, columns):
        self.name = name
        self.columns = pd.MultiIndex.from_tuples([("name", "units")] + columns)
        self.data_frame = pd.DataFrame([], columns=self.columns)
        self.row_list = []

    def add_data(self, data):
        self.row_list.append(data)

    def add_data_list(self, data):
        self.row_list.extend(data)

    def get_data_to_write(self):
        df = pd.DataFrame(self.row_list, columns=self.columns)
        df2 = self.data_frame.append(df)
        self.data_frame = df2.reset_index(drop=True)
        self.row_list = []
        return self.name, self.data_frame


class FileData:
    component_str = "_comp_"

    def __init__(self, calculation: BaseCalculation):
        """
        :param calculation:
        """
        self.file_path = calculation.measurement_file_path
        ext = path.splitext(calculation.measurement_file_path)[1]
        if ext == ".xlsx":
            self.file_type = FileType.excel_xlsx_file
        elif ext == ".xls":
            self.file_type = FileType.excel_xls_file
        else:
            self.file_type = FileType.text_file
        self.writing = False
        self.sheet_dict = dict()
        self.sheet_set = set()
        self.new_count = 0
        self.write_threshold = 40
        self.wrote_queue = Queue()
        self.error_queue = Queue()
        self.write_thread = threading.Thread(target=self.wrote_data_to_file)
        self.write_thread.daemon = True
        self.write_thread.start()
        self.add_data_part(calculation)

    def finished(self):
        """check if any data wait on write to disc"""
        return not self.writing and self.wrote_queue.empty()

    def good_sheet_name(self, name):
        if self.file_type == FileType.text_file:
            return False, "Text file allow store only one sheet"
        if FileData.component_str in name:
            return False, "Sequence '{}' is reserved for auto generated sheets".format(FileData.component_str)
        if name in self.sheet_set:
            return False, "Sheet name {} already in use".format(name)
        return True, True

    def add_data_part(self, calculation: BaseCalculation):
        """
        :type calculation: Calculation
        :param calculation:
        :return:
        """
        if calculation.measurement_file_path != self.file_path:
            raise ValueError(
                "[FileData] different file path {} vs {}".format(calculation.measurement_file_path, self.file_path)
            )
        if calculation.sheet_name in self.sheet_set:
            raise ValueError("[FileData] sheet name {} already in use".format(calculation.sheet_name))
        measurement = calculation.calculation_plan.get_measurements()
        component_information = [x.statistic_profile.get_component_info(x.units) for x in measurement]
        num = 1
        sheet_list = []
        header_list = []
        main_header = []
        for i, el in enumerate(component_information):
            local_header = []
            component_seg = False
            component_mask = False
            for per_component, area in measurement[i].statistic_profile.get_component_and_area_info():
                if per_component == PerComponent.Yes and area == AreaType.Segmentation:
                    component_seg = True
                if per_component == PerComponent.Yes and area != AreaType.Segmentation:
                    component_mask = True
            if component_seg:
                local_header.append(("Segmentation component", "num"))
            if component_mask:
                local_header.append(("Mask component", "num"))
            if any([x[1] for x in el]):
                sheet_list.append(
                    "{}{}{} - {}".format(
                        calculation.sheet_name,
                        FileData.component_str,
                        num,
                        measurement[i].name_prefix + measurement[i].name,
                    )
                )
                num += 1
            else:
                sheet_list.append(None)
            for name, comp in el:
                local_header.append(name)
                if not comp:
                    main_header.append(name)
            header_list.append(local_header)

        self.sheet_dict[calculation.uuid] = (
            SheetData(calculation.sheet_name, main_header),
            [SheetData(name, header_list[i]) if name is not None else None for i, name in enumerate(sheet_list)],
            component_information,
        )

    def wrote_data(self, uuid, data: ResponseData):
        self.new_count += 1
        main_sheet, component_sheets, _component_information = self.sheet_dict[uuid]
        name = data.path_to_file
        data_list = [name]
        for el, comp_sheet in zip(data.values, component_sheets):
            data_list.extend(el.get_global_parameters()[1:])
            comp_list = el.get_separated()
            if comp_sheet is not None:
                comp_sheet.add_data_list(comp_list)
        main_sheet.add_data(data_list)
        if self.new_count >= self.write_threshold:
            self.dump_data()
            self.new_count = 0

    def dump_data(self):
        data = []
        for main_sheet, component_sheets, _ in self.sheet_dict.values():
            data.append(main_sheet.get_data_to_write())
            for sheet in component_sheets:
                if sheet is not None:
                    data.append(sheet.get_data_to_write())
        self.wrote_queue.put(data)

    def wrote_data_to_file(self):
        while True:
            data = self.wrote_queue.get()
            self.writing = True
            if data == "finish":
                break
            try:
                if self.file_type == FileType.text_file:
                    base_path, ext = path.splitext(self.file_path)
                    for sheet_name, data_frame in data:
                        data_frame.to_csv(base_path + "_" + sheet_name + ext)
                else:
                    writer = pd.ExcelWriter(self.file_path)
                    new_sheet_names = []
                    ind = 0
                    for sheet_name, _ in data:
                        if len(sheet_name) < 32:
                            new_sheet_names.append(sheet_name)
                        else:
                            new_sheet_names.append(sheet_name[:27] + f"_{ind}_")
                            ind += 1
                    for sheet_name, (_, data_frame) in zip(new_sheet_names, data):
                        data_frame.to_excel(writer, sheet_name=sheet_name)
                    writer.save()
            except Exception as e:
                logging.error(e)
                self.error_queue.put(e)
            finally:
                self.writing = False

    def get_errors(self):
        res = []
        while not self.error_queue.empty():
            res.append(self.error_queue.get())
        return res

    def finish(self):
        self.wrote_queue.put("finish")

    def is_empty_sheet(self, sheet_name):
        return sheet_name not in self.sheet_set


class DataWriter:
    def __init__(self):
        self.file_dict: typing.Dict[str, FileData] = dict()

    def is_empty_sheet(self, file_path, sheet_name):
        if FileData.component_str in sheet_name:
            return False
        if file_path not in self.file_dict:
            return True
        return self.file_dict[file_path].is_empty_sheet(sheet_name)

    def add_data_part(self, calculation: BaseCalculation):
        if calculation.measurement_file_path in self.file_dict:
            self.file_dict[calculation.measurement_file_path].add_data_part(calculation)
        else:
            self.file_dict[calculation.measurement_file_path] = FileData(calculation)

    def add_result(self, data: ResponseData, calculation: BaseCalculation):
        if calculation.measurement_file_path not in self.file_dict:
            raise ValueError("Unknown measurement file")
        file_writer = self.file_dict[calculation.measurement_file_path]
        file_writer.wrote_data(calculation.uuid, data)
        return file_writer.get_errors()

    def writing_finished(self) -> bool:
        """check if all data are written to disc"""
        return all([x.finished() for x in self.file_dict.values()])

    def finish(self):
        for file_data in self.file_dict.values():
            file_data.finish()

    def calculation_finished(self, calculation):
        if calculation.measurement_file_path not in self.file_dict:
            raise ValueError("Unknown measurement file")
        self.file_dict[calculation.measurement_file_path].dump_data()
        return self.file_dict[calculation.measurement_file_path].get_errors()
